@@ -1,15 +1,24 @@
 from __future__ import annotations
+
+import pickle
+from pathlib import Path
 from functools import partial, wraps
 
+import torchaudio
+from torchaudio.functional import resample
+
+import sounddevice as sd
+
 import torch
 import torch.nn.functional as F
-from torch import nn, compiler, Tensor, tensor, is_tensor, cat, stft, istft, hann_window, view_as_real, view_as_complex
-from torch.nn import LSTM, GRU, Module, ModuleList
+from torch.fft import irfft
+from torch import nn, compiler, Tensor, tensor, is_tensor, cat, stft, hann_window, view_as_real, view_as_complex
+from torch.nn import LSTM, GRU, ConvTranspose1d, Module, ModuleList
 
 from numpy import ndarray
 
 from einx import add, multiply, divide
-from einops import rearrange, repeat, pack, unpack
+from einops import rearrange, repeat, reduce, pack, unpack
 from einops.layers.torch import Rearrange
 
 # ein tensor notation:
@@ -81,7 +90,7 @@ class STFT(Module):
         self.hop_length = hop_length
         self.win_length = win_length
 
-        window = torch.hann_window(win_length)
+        window = hann_window(win_length)
         self.register_buffer('window', window)
 
         self.eps = eps
@@ -94,7 +103,7 @@ class STFT(Module):
 
         # inverse FFT
 
-        ifft = torch.fft.irfft(spec, n_fft, dim = 1, norm = 'backward')
+        ifft = irfft(spec, n_fft, dim = 1, norm = 'backward')
 
         ifft = multiply('b w f, w', ifft, window)
 
@@ -141,23 +150,62 @@ class STFT(Module):
 
         return stft
 
+# Tasnet applies a hann window to the convtranspose1d
+
+class ConvTranspose1DWithHannWindow(ConvTranspose1d):
+    def __init__(
+        self,
+        dim,
+        dim_out,
+        filters,
+        **kwargs
+    ):
+        super().__init__(dim, dim_out, filters, **kwargs)
+
+        self.register_buffer('window', hann_window(filters))
+
+    def forward(self, x):
+
+        filters = self.weight
+
+        windowed_filters = multiply('o i k, k', filters, self.window)
+
+        return F.conv_transpose1d(
+            x,
+            windowed_filters,
+            stride = self.stride,
+            padding = self.padding
+        )
+
 # classes
 
 class HSTasNet(Module):
     def __init__(
         self,
-        dim = 500,          # they have 500 hidden units for the network, with 1000 at fusion (concat from both representation branches)
-        small = False,      # params cut in half by 1 layer lstm vs 2, fusion uses summed representation
+        *,
+        dim = 500,            # they have 500 hidden units for the network, with 1000 at fusion (concat from both representation branches)
+        small = False,        # params cut in half by 1 layer lstm vs 2, fusion uses summed representation
         stereo = False,
-        num_basis = 1024,
+        num_basis = 1500,
         segment_len = 1024,
         overlap_len = 512,
         n_fft = 1024,
-        num_sources = 4,    # drums, bass, vocals, other
+        sample_rate = 44_100, # they use 41k sample rate
+        num_sources = 4,      # drums, bass, vocals, other
         torch_compile = False,
         use_gru = False
     ):
         super().__init__()
+
+        # auto-saving config
+
+        _locals = locals()
+        _locals.pop('self', None)
+        _locals.pop('__class__', None)
+        self._config = pickle.dumps(_locals)
+
+        # hyperparameters
+
         audio_channels = 2 if stereo else 1
 
         self.audio_channels = audio_channels
@@ -171,6 +219,10 @@ class HSTasNet(Module):
         assert divisible_by(segment_len, 2)
         self.causal_pad = segment_len // 2
 
+        # sample rate - 41k in paper
+
+        self.sample_rate = sample_rate
+
         # spec branch encoder stft hparams
 
         self.stft = STFT(
@@ -207,7 +259,7 @@ class HSTasNet(Module):
             Rearrange('... (t basis) -> ... basis t', t = num_sources)
         )
 
-        self.conv_decode = nn.ConvTranspose1d(num_basis, audio_channels, segment_len, stride = overlap_len)
+        self.conv_decode = ConvTranspose1DWithHannWindow(num_basis, audio_channels, segment_len, stride = overlap_len)
 
         # they do a single layer of lstm in their "small" variant
 
@@ -233,7 +285,57 @@ class HSTasNet(Module):
         if torch_compile:
             self.forward = torch.compile(self.forward)
 
-    def init_stream_fn(self, device = None):
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
+    @property
+    def num_parameters(self):
+        return sum([p.numel() for p in self.parameters()])
+
+    # saving and loading
+
+    def save(self, path, overwrite = True):
+        path = Path(path)
+        assert overwrite or not path.exists()
+
+        pkg = dict(
+            model = self.state_dict(),
+            config = self._config,
+        )
+
+        torch.save(pkg, str(path))
+
+    def load(self, path, strict = True):
+        path = Path(path)
+        assert path.exists()
+
+        pkg = torch.load(str(path), map_location = 'cpu')
+
+        self.load_state_dict(pkg['model'], strict = strict)
+
+    @classmethod
+    def init_and_load_from(cls, path, strict = True):
+        path = Path(path)
+        assert path.exists()
+        pkg = torch.load(str(path), map_location = 'cpu')
+
+        assert 'config' in pkg, 'model configs were not found in this saved checkpoint'
+
+        config = pickle.loads(pkg['config'])
+        model = cls(**config)
+
+        self.load(model, path, strict = strict)
+        return model
+
+    # returns a function that closures the hiddens and past audio chunk for integration with sounddevice audio callback
+
+    def init_stateful_transform_fn(
+        self,
+        device = None,
+        return_reduced_sources: list[int] | None = None,
+    ):
+        chunk_len = self.overlap_len
         self.eval()
 
         past_audio = torch.zeros((self.audio_channels, self.overlap_len), device = device)
@@ -241,7 +343,7 @@ class HSTasNet(Module):
 
         @torch.inference_mode()
         def fn(audio_chunk: ndarray | Tensor):
-            assert audio_chunk.shape[-1] == self.overlap_len
+            assert audio_chunk.shape[-1] == chunk_len
 
             nonlocal hiddens
             nonlocal past_audio
@@ -267,12 +369,12 @@ class HSTasNet(Module):
 
             # forward chunk with past overlap through model
 
-            transformed, hiddens = self.forward(full_chunk, hiddens = hiddens)
+            transformed, hiddens = self.forward(full_chunk, hiddens = hiddens, return_reduced_sources = return_reduced_sources)
 
             transformed = rearrange(transformed, '1 ... -> ...')
 
             if squeezed_audio_channel:
-                transformer = rearrange(transformed, 't 1 ... -> t ...')
+                transformed = rearrange(transformed, '... 1 n -> ... n')
 
             if is_numpy_input:
                 transformed = transformed.cpu().numpy()
@@ -281,20 +383,110 @@ class HSTasNet(Module):
 
             past_audio = audio_chunk
 
-            return transformed
+            return transformed[..., -chunk_len:]
 
         return fn
 
-    @property
-    def num_parameters(self):
-        return sum([p.numel() for p in self.parameters()])
+    def sounddevice_stream(
+        self,
+        return_reduced_sources: list[int],
+        duration_seconds = 10,
+        channels = None,
+        device = None,
+        **stream_kwargs
+    ):
+        assert len(return_reduced_sources) > 0
+
+        transform_fn = self.init_stateful_transform_fn(
+            return_reduced_sources = return_reduced_sources,
+            device = device
+        )
+
+        # sounddevice stream callback, where raw audio can be transformed
+
+        def callback(indata, outdata, frames, time, status):
+            assert indata.shape[0] == self.overlap_len
+
+            indata = rearrange(indata, 'n c -> c n')
+
+            transformed = transform_fn(indata)
+
+            transformed = rearrange(transformed, 'c n -> n c')
+
+            outdata[:] = transformed
+
+        # variables
+
+        duration_ms = int(duration_seconds * 1000)
+
+        # sounddevice streaming
+
+        with sd.Stream(
+            **stream_kwargs,
+            channels = channels,
+            callback = callback,
+            samplerate = self.sample_rate
+        ):
+            sd.sleep(duration_ms)
+
+    def process_audio_file(
+        self,
+        input_file: str | Path,
+        return_reduced_sources: list[int],
+        output_file: str | Path | None = None,
+        overwrite = False
+    ):
+        if isinstance(input_file, str):
+            input_file = Path(input_file)
+
+        assert len(return_reduced_sources) > 0
+        assert input_file.exists(), f'{str(input_file)} not found'
+
+        audio_tensor, sample_rate = torchaudio.load(input_file)
+
+        # resample if need be
+
+        if sample_rate != self.sample_rate:
+            audio_tensor = resample(audio_tensor, sample_rate, self.sample_rate)
+
+        # curtail to divisible segment lens
+
+        audio_len = audio_tensor.shape[-1]
+        rounded_down_len = audio_len // self.segment_len * self.segment_len
+
+        audio_tensor = audio_tensor[..., :rounded_down_len]
+
+        # transform
+
+        audio_tensor = audio_tensor.to(self.device)
+
+        with torch.no_grad():
+            self.eval()
+            transformed = self.forward(audio_tensor, return_reduced_sources = return_reduced_sources)
+
+        # save output file
+
+        if not exists(output_file):
+            output_file = Path(input_file.parents[-2] / f'{input_file.stem}-out.mp3')
+
+        assert output_file != input_file
+
+        assert not output_file.exists() or overwrite, f'{str(output_file)} already exists, set `overwrite = True` if you wish to overwrite'
+
+        torchaudio.save(str(output_file), audio_tensor.cpu(), sample_rate = self.sample_rate)
+
+        print(f'separated audio saved to {str(output_file)}')
 
     def forward(
         self,
         audio,
         hiddens = None,
-        targets = None
+        targets = None,
+        return_reduced_sources: list[int] | None = None,
+        auto_causal_pad = None
     ):
+        auto_causal_pad = default(auto_causal_pad, self.training)
+
         batch, audio_len, device = audio.shape[0], audio.shape[-1], audio.device
 
         assert divisible_by(audio_len, self.segment_len)
@@ -315,14 +507,13 @@ class HSTasNet(Module):
 
         # pad the audio manually on the left side for causal, and set stft center False
 
-        audio = F.pad(audio, (self.causal_pad, 0), value = 0.)
+        if auto_causal_pad:
+            audio = F.pad(audio, (self.causal_pad, 0), value = 0.)
 
         # handle spec encoding
 
         spec_audio_input = rearrange(audio, 'b s ... -> (b s) ...')
 
-        stft_window = hann_window(self.segment_len, device = device)
-
         complex_spec = self.stft(spec_audio_input)
 
         real_spec = view_as_real(complex_spec)
@@ -424,7 +615,8 @@ class HSTasNet(Module):
 
         # excise out the causal padding
 
-        recon_audio = recon_audio[..., self.causal_pad:]
+        if auto_causal_pad:
+            recon_audio = recon_audio[..., self.causal_pad:]
 
         if exists(targets):
             recon_loss = F.l1_loss(recon_audio, targets) # they claim a simple l1 loss is better than all the complicated stuff of past
@@ -440,4 +632,10 @@ class HSTasNet(Module):
             next_post_waveform_hidden
         )
 
+        if exists(return_reduced_sources):
+            return_reduced_sources = tensor(return_reduced_sources, device = device)
+
+            sources = recon_audio.index_select(1, return_reduced_sources)
+            recon_audio = reduce(sources, 'b s ... -> b ...', 'sum')
+
         return recon_audio, lstm_hiddens
